#!/usr/bin/python 
###########################################################
#
# Copyright (c) 2005, Southpaw Technology
#                     All Rights Reserved
#
# PROPRIETARY INFORMATION.  This software is proprietary to
# Southpaw Technology, and is not to be reproduced, transmitted,
# or disclosed in any way without written permission.
#
#
#
import tacticenv

from pyasm.security import *
from transaction import *
from search import *
from sql import *

import unittest

class SqlTest(unittest.TestCase):


    def test_all(my):
        my._test_get_connect()
        my._test_select_class()
        my._test_insert_class()
        my._test_update_class()
        my._test_insert_and_delete()
        my._test_create_table()
        my._test_transaction()
        my._test_order_by()
        my._test_rpn_filters()
        my._test_search_filter()


    def _test_get_connect(my):
        sql1 = DbContainer.get("unittest")
        sql2 = DbContainer.get("unittest")
        my.assertEquals(sql1, sql2)

    def _test_select_class(my):
        """ test a select """
        select = Select()
        select.set_database('unittest')
        select.add_table("person")
        select.add_where("\"name_first\" = 'megumi'")
        select.add_order_by("name_last")

        statement = select.get_statement()
        expected = "SELECT * FROM \"person\" WHERE \"name_first\" = 'megumi' ORDER BY \"person\".\"name_last\""

        my.assertEquals( expected, statement )

        select = Select()
        select.set_database('unittest')
        select.add_table("person")
        select.add_filter('name_last', "john's", op='!=')
        statement = select.get_statement()
        expected = "SELECT * FROM \"person\" WHERE \"name_last\" != 'john''s'"
        my.assertEquals( expected, statement )


    def _test_insert_class(my):
        """test an insert"""
        insert = Insert()
        insert.set_table("person");
        insert.set_value("name_first", "megumi");
        insert.set_value("name_last", "takamori");
        statement = insert.get_statement()

        sql = DbContainer.get("unittest")

        if sql.get_database_type() == "Oracle":
            expected = "INSERT INTO \"person\" (\"id\", \"name_first\", \"name_last\") VALUES (\"person_id_seq\".nextval, 'megumi', 'takamori')"
        else:
            expected = "INSERT INTO \"person\" (\"name_first\", \"name_last\") VALUES ('megumi', 'takamori')"

        my.assertEquals( expected, statement )


    def _test_update_class(my):
        """test an update"""
        update = Update()
        update.set_table("person");
        update.set_value("name_first", "megumi");
        update.add_where("\"person_id\" = '1'");
        statement = update.get_statement()
        expected = "UPDATE \"person\" SET \"name_first\" = 'megumi' WHERE \"person_id\" = '1'"

        my.assertEqual( expected, statement )



    def _test_insert_and_delete(my):

        # ensure that we are *NOT in a transaction
        transaction = Transaction.get()
        my.assertEquals( None, transaction )

        sql = DbContainer.get("unittest")
        my.assertEquals( False, sql.is_in_transaction() )


        count_sql = """select count(*) from "person"
                    where "name_first" = 'Bugs' and "name_last" = 'Bunny'"""

        num_records = sql.get_int(count_sql)
        my.assertEquals(0, num_records)

        # test with no transaction
        transaction = Transaction.get(create=True)

        insert = Insert()
        insert.set_table("person")
        insert.set_value("name_first", "Bugs")
        insert.set_value("name_last", "Bunny")
        statement = insert.get_statement()
        sql.do_update(statement)

        num_records = sql.get_int(count_sql)
        my.assertEquals(1, num_records)

        delete = """delete from "person"
                 where "name_first" = 'Bugs' and "name_last" = 'Bunny'"""
        sql.do_update(delete)

        num_records = sql.get_int(count_sql)
        my.assertEquals(0, num_records)

        transaction.rollback()



    def _test_create_table(my):

        create = CreateTable()
        create.set_table("coffee")

        create.add_column("id", "int4")
        create.add_column("type", "varchar(10)")
        create.add_column("login", "varchar(30)")
        create.add_column("discussion", "text")

        create.set_primary_key("id")

        statement = create.get_statement()
        sql = DbContainer.get("unittest")
        if sql.get_database_type() == "Oracle":
            expected = \
'''CREATE TABLE "coffee" (
    "id" NUMBER,
    "type" VARCHAR2(10),
    "login" VARCHAR2(30),
    "discussion" CLOB,
    PRIMARY KEY ("id")
);'''
        else:
            expected = \
'''CREATE TABLE "coffee" (
    "id" int4,
    "type" varchar(10),
    "login" varchar(30),
    "discussion" text,
    PRIMARY KEY ("id"));'''



        statement = statement.replace("\n", "")
        statement = statement.replace("    ", " ")
        statement = statement.replace("\t", " ")
        expected = expected.replace("\n", "")
        statement = expected.replace("    ", " ")
        statement = expected.replace("\t", " ")

        my.assertEquals(expected, statement)




    def _test_transaction(my):
        """test a transaction"""
        sql = DbContainer.get("unittest")

        count_sql = 'select count(*) from "person"'

        # start the transaction, update and roll back
        sql.start()

        num_records = sql.get_int(count_sql)

        insert = Insert()
        insert.set_table("person")
        insert.set_value("name_first", "cow")
        insert.set_value("name_last", "sql")
        query = insert.get_statement()

        sql.do_update(query)
        new_num_records = sql.get_value(count_sql)
        my.assertEquals( new_num_records, num_records+1 )
        sql.rollback()

        # dump after the rollback
        new_num_records = sql.get_int(count_sql)
        my.assertEqual( new_num_records, num_records )



    def _test_order_by(my):

        select = Select()
        select.add_table("asset")
        select.add_enum_order_by("code", ['cow', 'dog', 'horse'])

        expected = '''SELECT * FROM "asset" ORDER BY ( CASE "code"
WHEN 'cow' THEN 1 
WHEN 'dog' THEN 2 
WHEN 'horse' THEN 3 
ELSE 4 END )'''

        statement = select.get_statement()

        my.assertEqual(expected, statement)




    def _test_rpn_filters(my):
        select = Select()
        select.add_table("asset")
        select.add_where("begin")
        select.add_where("\"code\" = 'chr001'")
        select.add_where("\"code\" = 'chr002'")
        select.add_where("\"code\" = 'chr003'")
        select.add_where("or")
        select.add_where("\"status\" = 'complete'")
        select.add_where("and")
        statement = select.get_statement()

        expected = '''SELECT * FROM "asset" WHERE ( ( "code" = 'chr001' OR "code" = 'chr002' OR "code" = 'chr003' ) AND "status" = 'complete' )'''
        my.assertEquals(expected, statement)



        # test some simple fringe cases
        select = Select()
        select.add_table("asset")
        select.add_where("begin")
        select.add_where("or")
        select.add_where("\"status\" = 'complete'")
        select.add_where("and")
        statement = select.get_statement()
        expected = "SELECT * FROM \"asset\" WHERE \"status\" = 'complete'"
        my.assertEquals(expected, statement)

 

        select = Select()
        select.add_table("asset")
        select.add_where("\"status\" = 'retired'")
        select.add_where("\"code\" = 'chr001'")
        select.add_where("\"code\" = 'chr002'")
        select.add_where("or")
        statement = select.get_statement()
        expected = '''SELECT * FROM "asset" WHERE "status" = 'retired' AND ( "code" = 'chr001' OR "code" = 'chr002' )'''
        my.assertEquals(expected, statement)



        # add a more complex case
        search = Select()
        search.set_database('unittest')
        search.add_table("person")
        search.add_filter("login", "joe")
        search.add_filter("login", "mary")
        search.add_where("and")
        search.add_filter("attr", "tom")
        search.add_filter("attr", "peter")
        search.add_where("and")

        search.add_where("or")
        statement = search.get_statement()
        expected = '''SELECT * FROM "person" WHERE ( ( "login" = 'joe' AND "login" = 'mary' ) OR ( "attr" = 'tom' AND "attr" = 'peter' ) )'''

        my.assertEquals(expected, statement)



    def _test_search_filter(my):

        select = Select()
        select.set_database('unittest')
        select.add_table("job")
        select.add_column("request_id")
        select.add_filter("code", "123MMS")

        select2 = Select()
        select2.set_database('unittest')
        select2.add_table("request")
        select2.add_select_filter("id", select)
        statement = select2.get_statement()
        expected = '''SELECT * FROM "request" WHERE "id" in ( SELECT "job"."request_id" FROM "job" WHERE "code" = '123MMS' )'''
        my.assertEquals(expected, statement)

        select3 = Select()
        select3.set_database('unittest')
        select3.add_op("begin")
        select3.add_table("request")
        select3.add_select_filter("id", select)

        statement = select3.get_statement()
        expected = '''SELECT * FROM "request" WHERE "id" in ( SELECT "job"."request_id" FROM "job" WHERE "code" = '123MMS' )'''
        my.assertEquals(expected, statement)
 



if __name__ == "__main__":
    Batch()
    unittest.main()

